import { enablePatches, produce } from 'immer';
import { createContext, useCallback, useContext, useEffect, useMemo, useState, } from 'react';
import { devtools, subscribeWithSelector } from 'zustand/middleware';
import { createWithEqualityFn } from 'zustand/traditional';
import isDeepEqual from 'fast-deep-equal';
import { shallow } from 'zustand/shallow';
import { converter, historyManagement, immerWithPatches, } from './middleware.js';
import { applyFlatConfigToState, createQueryObject, flattenConfig, } from './utils.js';
enablePatches();
export const DEFAULT_NAMESPACE = 'default';
export const StoreContext = createContext(null);
export const createGeschichte = (historyInstance) => {
    const thisStore = converter(historyInstance);
    const storeWithHistory = historyManagement(historyInstance)(thisStore);
    const middleware = immerWithPatches(storeWithHistory);
    if (process.env.NODE_ENV !== 'production' && typeof window !== 'undefined') {
        return createWithEqualityFn(subscribeWithSelector(devtools(middleware, { name: 'geschichte' })), shallow);
    }
    return createWithEqualityFn(subscribeWithSelector(middleware), shallow);
};
const assertContextExists = (value) => {
    if (process.env.NODE_ENV !== 'production' && value === null) {
        throw new Error('Cannot find `GeschichteProvider` in React tree context. Please provide outer <GeschichteProvider />.');
    }
    return value;
};
export const useStore = () => {
    return assertContextExists(useContext(StoreContext));
};
export const useBatchQuery = () => {
    const store = useStore();
    return store(({ batchPushState, batchReplaceState }) => ({
        batchPushState,
        batchReplaceState,
    }));
};
export const factoryParameters = (config, defaultInitialValues = {}, ns = DEFAULT_NAMESPACE) => {
    const flatConfig = flattenConfig(config);
    const createInitialValues = (d) => typeof d === 'function' ? d() : d;
    const initBlank = (initialQueries, initialValues) => {
        // thisValues will be mutated by applyFlatConfigToState, that's why we init it with a copy of
        // the initial state.
        let thisQuery = {};
        // We produce a new state here instead of mutating defaultInitialValues.
        // Otherwise it's possible that it get's reused across executions and that will yield to readonly errors.
        const values = produce(initialValues, (draft) => {
            thisQuery = applyFlatConfigToState(flatConfig, initialQueries, ns, draft, initialValues);
        });
        return {
            initialValues,
            query: thisQuery,
            values,
        };
    };
    const useQuery = () => {
        const useStore = assertContextExists(useContext(StoreContext));
        const { register, pushState, replaceState, resetPush, resetReplace, initialQueries, } = useStore(({ register, pushState, replaceState, resetPush, resetReplace, initialQueries, }) => ({
            initialQueries,
            pushState,
            register,
            replaceState,
            resetPush,
            resetReplace,
        }), shallow);
        const initialRegisterState = useMemo(() => {
            const initialValues = createInitialValues(defaultInitialValues);
            return initBlank(initialQueries(), initialValues);
            // eslint-disable-next-line react-hooks/exhaustive-deps
        }, [useStore, defaultInitialValues]);
        const [currentState, setCurrentState] = useState({
            initialValues: initialRegisterState.initialValues,
            values: initialRegisterState.values,
        });
        useEffect(() => {
            const { unsubscribe: unregister, values, initialValues, } = register(config, flatConfig, ns, initialRegisterState.initialValues, initialRegisterState.query, initialRegisterState.values);
            if (!isDeepEqual(initialRegisterState.values, values) ||
                !isDeepEqual(initialRegisterState.initialValues, initialValues) ||
                !isDeepEqual(currentState.initialValues, initialValues) ||
                !isDeepEqual(currentState.values, values)) {
                setCurrentState({ values, initialValues });
            }
            const unsubscribe = useStore.subscribe((state) => state.namespaces[ns] && {
                initialValues: state.namespaces[ns].initialValues,
                values: state.namespaces[ns].values,
            }, (state) => {
                if (state) {
                    setCurrentState(state);
                }
            }, { equalityFn: shallow });
            return () => {
                unsubscribe();
                unregister();
            };
            // eslint-disable-next-line react-hooks/exhaustive-deps
        }, [initialRegisterState]);
        const values = currentState.values;
        const initialValues = currentState.initialValues;
        const createQuery = useCallback((values) => {
            return createQueryObject(flatConfig, ns, values, initialValues);
        }, [initialValues]);
        return useMemo(() => ({
            createQuery: (customValues) => createQuery(customValues || values),
            createQueryString: (customValues) => new URLSearchParams(createQuery(customValues || values)).toString(),
            initialValues,
            pushState: (state, options) => pushState(ns, state, options),
            replaceState: (state, options) => replaceState(ns, state, options),
            resetPush: () => resetPush(ns),
            resetReplace: () => resetReplace(ns),
            values,
        }), [
            values,
            initialValues,
            pushState,
            replaceState,
            resetPush,
            resetReplace,
            createQuery,
        ]);
    };
    const createQueryString = (values, initialValues) => {
        const thisInitialValues = typeof initialValues === 'undefined'
            ? createInitialValues(defaultInitialValues)
            : initialValues;
        return new URLSearchParams(createQueryObject(flatConfig, ns, values, thisInitialValues)).toString();
    };
    const parseQueryString = (query, initialValues) => {
        const thisInitialValues = typeof initialValues === 'undefined'
            ? createInitialValues(defaultInitialValues)
            : initialValues;
        return produce({}, (draft) => {
            const parsedQuery = new URLSearchParams(query);
            applyFlatConfigToState(flatConfig, Object.fromEntries(parsedQuery.entries()), ns, draft, thisInitialValues);
        });
    };
    return { useQuery, createQueryString, parseQueryString };
};
